/****************************************************************************
**  CUBE        http://www.score-p.org/                                    **
**  SCALASCA    http://www.scalasca.org/                                   **
*****************************************************************************
**  Copyright (c) 1998-2014                                                **
**  Forschungszentrum Juelich GmbH, Juelich Supercomputing Centre          **
**                                                                         **
**  Copyright (c) 2009-2014                                                **
**  German Research School for Simulation Sciences GmbH,                   **
**  Laboratory for Parallel Programming                                    **
**                                                                         **
**  This software may be modified and distributed under the terms of       **
**  a BSD-style license.  See the COPYING file in the package base         **
**  directory for details.                                                 **
****************************************************************************/

/**
 * \file CubePreDerivedInclusiveMetric.cpp
 * \brief Defines methods to calculate incl/exclusve values if the metric is defined as an expression. Expression gets calculated first, then the value gets aggregated.

 ********************************************/


#include <iostream>

#include "CubePreDerivedInclusiveMetric.h"

using namespace std;
using namespace cube;


PreDerivedInclusiveMetric::~PreDerivedInclusiveMetric()
{
}

// inclusive value and exclusive value in this metric are same
double
PreDerivedInclusiveMetric::get_sev( Cnode* cnode, CalculationFlavour cnf, Sysres* sys, CalculationFlavour sf )
{
    pre_calculation_preparation( cnode, sys );
    if ( init_evaluation != NULL )
    {
        init_evaluation->eval();
        delete init_evaluation;
        init_evaluation = NULL;
    }

    Value* cached_value = cache->getCachedValue( cnode, cnf, sys, sf );
    if ( cached_value != NULL )
    {
	double _d = cached_value->getDouble();
	delete cached_value;
        return _d;
    }

    double v =  0.;

    if ( evaluation != NULL )
    {
        if ( ( sys->isSystemTreeNode() || sys->isLocationGroup() ) && sf == cube::CUBE_CALCULATE_EXCLUSIVE )
        {
            post_calculation_cleanup();
            return v;
        }
        if ( sys->isSystemTreeNode() )
        {
            SystemTreeNode* _sys = ( SystemTreeNode* )sys;
            // first add values of all sub system nodes
            for ( unsigned i = 0; i < _sys->num_children(); ++i )
            {
                double _v = get_sev( cnode, cnf, _sys->get_child( i ), cube::CUBE_CALCULATE_INCLUSIVE );
                ( v ) += _v;
            }
            // then add all values of all sub local groups
            for ( unsigned i = 0; i < _sys->num_groups(); ++i )
            {
                double _v = get_sev( cnode, cnf, _sys->get_location_group( i ), cube::CUBE_CALCULATE_INCLUSIVE );
                ( v ) += _v;
            }
        }
        if ( sys->isLocationGroup() )
        {
            LocationGroup* _lg = ( LocationGroup* )sys;
            // first add values of all locations
            for ( unsigned i = 0; i < _lg->num_children(); ++i )
            {
                double _v = get_sev( cnode, cnf, _lg->get_child( i ), cube::CUBE_CALCULATE_INCLUSIVE );
                ( v ) += _v;
            }
        }
        if ( sys->isLocation() )
        {
            Location* _loc   = ( Location* )sys;
            double    _value = evaluation->eval( cnode, CUBE_CALCULATE_INCLUSIVE, _loc, sf );
            ( v ) = _value;
            if ( cnf == CUBE_CALCULATE_EXCLUSIVE )
            {
                for ( cnode_id_t cid = 0; cid < cnode->num_children(); ++cid )
                {
                    if ( cnode->get_child( cid )->isVisible() ) // add only visible children, coz the hidden have been added in excl value
                    {
                        double _v = get_sev(  cnode->get_child( cid ), CUBE_CALCULATE_INCLUSIVE, _loc, cube::CUBE_CALCULATE_INCLUSIVE   );
                        ( v ) -= _v;
                    }
                }
            }
        }
    }
     Value* _v  =  metric_value->clone();
     (*_v) = v;
    cache->setCachedValue( _v, cnode, cnf, sys, sf );
    delete _v;
    post_calculation_cleanup();
    return v;
}




// inclusive value and exclusive value in this metric are same
Value*
PreDerivedInclusiveMetric::get_sev_adv( Cnode* cnode, CalculationFlavour cnf, Sysres* sys, CalculationFlavour sf )
{
     Value* _v  =  metric_value->clone();
    (*_v) = get_sev( cnode, cnf, sys, sf);
    return _v;
}



double
PreDerivedInclusiveMetric::get_sev( Cnode* cnode, CalculationFlavour cnf )
{
    pre_calculation_preparation( cnode );
    if ( init_evaluation != NULL )
    {
        init_evaluation->eval();
        delete init_evaluation;
        init_evaluation = NULL;
    }
    Value* cached_value = cache->getCachedValue( cnode, cnf );
    if ( cached_value != NULL )
    {
	double _d = cached_value->getDouble();
	delete cached_value;
        return _d;
    }

    double v = 0.;
    for ( std::vector<Location*>::iterator iter = sysv.begin(); iter != sysv.end(); ++iter )
    {  
	Location* _loc   = *iter;
	pre_calculation_preparation( cnode, _loc );
	double    _value = evaluation->eval( cnode, CUBE_CALCULATE_INCLUSIVE, _loc, CUBE_CALCULATE_EXCLUSIVE );
	v  += _value;
	post_calculation_cleanup();
    }

    if ( cnf == CUBE_CALCULATE_EXCLUSIVE )
    {
	for ( cnode_id_t cid = 0; cid < cnode->num_children(); ++cid )
	{
	    if ( cnode->get_child( cid )->isVisible() ) // add only visible children, coz the hidden have been added in excl value
	    {
		double _v = get_sev(  cnode->get_child( cid ), CUBE_CALCULATE_INCLUSIVE);
		( v ) -= _v;
	    }
	}
    }
 
    
    Value* _v  =  metric_value->clone();
    (*_v) = v;
    cache->setCachedValue( _v, cnode, cnf );
    delete _v;
    post_calculation_cleanup();
    return v;
}



Value*
PreDerivedInclusiveMetric::get_sev_adv( Cnode* cnode, CalculationFlavour cnf )
{
    Value* _v  =  metric_value->clone();
    (*_v) = get_sev( cnode, cnf);
    return _v;
}
